@@ -1,12 +1,17 @@
-import contextlib
 import math
 import warnings
 from types import ModuleType
+from typing import Any, cast
 
+import hypothesis
+import hypothesis.extra.numpy as npst
 import numpy as np
 import pytest
+from hypothesis import given
+from hypothesis import strategies as st
 
 from array_api_extra import (
+    apply_where,
     at,
     atleast_nd,
     broadcast_shapes,
@@ -23,27 +28,241 @@ from array_api_extra import (
 from array_api_extra._lib import Backend
 from array_api_extra._lib._testing import xp_assert_close, xp_assert_equal
 from array_api_extra._lib._utils._compat import device as get_device
+from array_api_extra._lib._utils._helpers import asarrays, eager_shape, ndindex
 from array_api_extra._lib._utils._typing import Array, Device
 from array_api_extra.testing import lazy_xp_function
 
 # some xp backends are untyped
 # mypy: disable-error-code=no-untyped-def
 
+lazy_xp_function(apply_where, static_argnums=(2, 3), static_argnames="xp")
 lazy_xp_function(atleast_nd, static_argnames=("ndim", "xp"))
 lazy_xp_function(cov, static_argnames="xp")
-# FIXME .device attribute https://github.com/data-apis/array-api-compat/pull/238
-lazy_xp_function(create_diagonal, jax_jit=False, static_argnames=("offset", "xp"))
+lazy_xp_function(create_diagonal, static_argnames=("offset", "xp"))
 lazy_xp_function(expand_dims, static_argnames=("axis", "xp"))
 lazy_xp_function(kron, static_argnames="xp")
 lazy_xp_function(nunique, static_argnames="xp")
 lazy_xp_function(pad, static_argnames=("pad_width", "mode", "constant_values", "xp"))
 # FIXME calls in1d which calls xp.unique_values without size
 lazy_xp_function(setdiff1d, jax_jit=False, static_argnames=("assume_unique", "xp"))
-# FIXME .device attribute https://github.com/data-apis/array-api-compat/pull/238
-lazy_xp_function(sinc, jax_jit=False, static_argnames="xp")
+lazy_xp_function(sinc, static_argnames="xp")
+
+
+@pytest.mark.skip_xp_backend(
+    Backend.SPARSE, reason="read-only backend without .at support"
+)
+class TestApplyWhere:
+    @staticmethod
+    def f1(x: Array, y: Array | int = 10) -> Array:
+        return x + y
+
+    @staticmethod
+    def f2(x: Array, y: Array | int = 10) -> Array:
+        return x - y
+
+    def test_f1_f2(self, xp: ModuleType):
+        x = xp.asarray([1, 2, 3, 4])
+        cond = x % 2 == 0
+        actual = apply_where(cond, x, self.f1, self.f2)
+        expect = xp.where(cond, self.f1(x), self.f2(x))
+        xp_assert_equal(actual, expect)
+
+    def test_fill_value(self, xp: ModuleType):
+        x = xp.asarray([1, 2, 3, 4])
+        cond = x % 2 == 0
+        actual = apply_where(x % 2 == 0, x, self.f1, fill_value=0)
+        expect = xp.where(cond, self.f1(x), xp.asarray(0))
+        xp_assert_equal(actual, expect)
+
+        actual = apply_where(x % 2 == 0, x, self.f1, fill_value=xp.asarray(0))
+        xp_assert_equal(actual, expect)
+
+    def test_args_tuple(self, xp: ModuleType):
+        x = xp.asarray([1, 2, 3, 4])
+        y = xp.asarray([10, 20, 30, 40])
+        cond = x % 2 == 0
+        actual = apply_where(cond, (x, y), self.f1, self.f2)
+        expect = xp.where(cond, self.f1(x, y), self.f2(x, y))
+        xp_assert_equal(actual, expect)
+
+    def test_broadcast(self, xp: ModuleType):
+        x = xp.asarray([1, 2])
+        y = xp.asarray([[10], [20], [30]])
+        cond = xp.broadcast_to(xp.asarray(True), (4, 1, 1))
+
+        actual = apply_where(cond, (x, y), self.f1, self.f2)
+        expect = xp.where(cond, self.f1(x, y), self.f2(x, y))
+        xp_assert_equal(actual, expect)
+
+        actual = apply_where(
+            cond,
+            (x, y),
+            lambda x, _: x,  # pyright: ignore[reportUnknownArgumentType]
+            lambda _, y: y,  # pyright: ignore[reportUnknownArgumentType]
+        )
+        expect = xp.where(cond, x, y)
+        xp_assert_equal(actual, expect)
+
+        # Shaped fill_value
+        actual = apply_where(cond, x, self.f1, fill_value=y)
+        expect = xp.where(cond, self.f1(x), y)
+        xp_assert_equal(actual, expect)
+
+    def test_dtype_propagation(self, xp: ModuleType, library: Backend):
+        x = xp.asarray([1, 2], dtype=xp.int8)
+        y = xp.asarray([3, 4], dtype=xp.int16)
+        cond = x % 2 == 0
+
+        mxp = np if library is Backend.DASK else xp
+        actual = apply_where(
+            cond,
+            (x, y),
+            self.f1,
+            lambda x, y: mxp.astype(x - y, xp.int64),  # pyright: ignore[reportArgumentType,reportUnknownArgumentType]
+        )
+        assert actual.dtype == xp.int64
+
+        actual = apply_where(cond, y, self.f1, fill_value=5)
+        assert actual.dtype == xp.int16
+
+    @pytest.mark.parametrize("fill_value_raw", [3, [3, 4]])
+    @pytest.mark.parametrize(
+        ("fill_value_dtype", "expect_dtype"), [("int32", "int32"), ("int8", "int16")]
+    )
+    def test_dtype_propagation_fill_value(
+        self,
+        xp: ModuleType,
+        fill_value_raw: int | list[int],
+        fill_value_dtype: str,
+        expect_dtype: str,
+    ):
+        x = xp.asarray([1, 2], dtype=xp.int16)
+        cond = x % 2 == 0
+        fill_value = xp.asarray(fill_value_raw, dtype=getattr(xp, fill_value_dtype))
+
+        actual = apply_where(cond, x, self.f1, fill_value=fill_value)
+        assert actual.dtype == getattr(xp, expect_dtype)
+
+    def test_dont_overwrite_fill_value(self, xp: ModuleType):
+        x = xp.asarray([1, 2])
+        fill_value = xp.asarray([100, 200])
+        actual = apply_where(x % 2 == 0, x, self.f1, fill_value=fill_value)
+        xp_assert_equal(actual, xp.asarray([100, 12]))
+        xp_assert_equal(fill_value, xp.asarray([100, 200]))
+
+    def test_dont_run_on_false(self, xp: ModuleType):
+        x = xp.asarray([1.0, 2.0, 0.0])
+        y = xp.asarray([0.0, 3.0, 4.0])
+        # On NumPy, division by zero will trigger warnings
+        actual = apply_where(
+            x == 0,
+            (x, y),
+            lambda x, y: x / y,  # pyright: ignore[reportUnknownArgumentType]
+            lambda x, y: y / x,  # pyright: ignore[reportUnknownArgumentType]
+        )
+        xp_assert_equal(actual, xp.asarray([0.0, 1.5, 0.0]))
+
+    def test_bad_args(self, xp: ModuleType):
+        x = xp.asarray([1, 2, 3, 4])
+        cond = x % 2 == 0
+        # Neither f2 nor fill_value
+        with pytest.raises(TypeError, match="Exactly one of"):
+            apply_where(cond, x, self.f1)  # type: ignore[call-overload]  # pyright: ignore[reportCallIssue]
+        # Both f2 and fill_value
+        with pytest.raises(TypeError, match="Exactly one of"):
+            apply_where(cond, x, self.f1, self.f2, fill_value=0)  # type: ignore[call-overload]  # pyright: ignore[reportCallIssue]
+
+    @pytest.mark.skip_xp_backend(Backend.NUMPY_READONLY, reason="xp=xp")
+    def test_xp(self, xp: ModuleType):
+        x = xp.asarray([1, 2, 3, 4])
+        cond = x % 2 == 0
+        actual = apply_where(cond, x, self.f1, self.f2, xp=xp)
+        expect = xp.where(cond, self.f1(x), self.f2(x))
+        xp_assert_equal(actual, expect)
+
+    def test_device(self, xp: ModuleType, device: Device):
+        x = xp.asarray([1, 2, 3, 4], device=device)
+        y = apply_where(x % 2 == 0, x, self.f1, self.f2)
+        assert get_device(y) == device
+        y = apply_where(x % 2 == 0, x, self.f1, fill_value=0)
+        assert get_device(y) == device
+        y = apply_where(x % 2 == 0, x, self.f1, fill_value=x)
+        assert get_device(y) == device
+
+    @pytest.mark.filterwarnings("ignore::RuntimeWarning")  # overflows, etc.
+    @hypothesis.settings(  # pyright: ignore[reportArgumentType]
+        # The xp and library fixtures are not regenerated between hypothesis iterations
+        suppress_health_check=[hypothesis.HealthCheck.function_scoped_fixture],
+        # JAX can take a long time to initialize on the first call
+        deadline=None,
+    )
+    @given(
+        n_arrays=st.integers(min_value=1, max_value=3),
+        rng_seed=st.integers(min_value=1000000000, max_value=9999999999),
+        dtype=st.sampled_from((np.float32, np.float64)),
+        p=st.floats(min_value=0, max_value=1),
+        data=st.data(),
+    )
+    def test_hypothesis(  # type: ignore[explicit-any,decorated-any]
+        self,
+        n_arrays: int,
+        rng_seed: int,
+        dtype: np.dtype[Any],
+        p: float,
+        data: st.DataObject,
+        xp: ModuleType,
+        library: Backend,
+    ):
+        mbs = npst.mutually_broadcastable_shapes(num_shapes=n_arrays + 1, min_side=0)
+        input_shapes, _ = data.draw(mbs)
+        cond_shape, *shapes = input_shapes
+
+        # cupy/cupy#8382
+        # https://github.com/jax-ml/jax/issues/26658
+        elements = {"allow_subnormal": library not in (Backend.CUPY, Backend.JAX)}
+
+        fill_value = xp.asarray(
+            data.draw(npst.arrays(dtype=dtype, shape=(), elements=elements))
+        )
+        float_fill_value = float(fill_value)
+        if library is Backend.CUPY and dtype is np.float32:
+            # Avoid data-dependent dtype promotion when encountering subnormals
+            # close to the max float32 value
+            float_fill_value = float(np.clip(float_fill_value, -1e38, 1e38))
+
+        arrays = tuple(
+            xp.asarray(
+                data.draw(npst.arrays(dtype=dtype, shape=shape, elements=elements))
+            )
+            for shape in shapes
+        )
+
+        def f1(*args: Array) -> Array:
+            return cast(Array, sum(args))
+
+        def f2(*args: Array) -> Array:
+            return cast(Array, sum(args) / 2)
+
+        rng = np.random.default_rng(rng_seed)
+        cond = xp.asarray(rng.random(size=cond_shape) > p)
+
+        res1 = apply_where(cond, arrays, f1, fill_value=fill_value)
+        res2 = apply_where(cond, arrays, f1, f2)
+        res3 = apply_where(cond, arrays, f1, fill_value=float_fill_value)
+
+        ref1 = xp.where(cond, f1(*arrays), fill_value)
+        ref2 = xp.where(cond, f1(*arrays), f2(*arrays))
+        if library is Backend.ARRAY_API_STRICT:
+            # FIXME https://github.com/data-apis/array-api-strict/issues/131
+            ref3 = xp.where(cond, *asarrays(f1(*arrays), float_fill_value, xp=xp))
+        else:
+            ref3 = xp.where(cond, f1(*arrays), float_fill_value)
+
+        xp_assert_close(res1, ref1, rtol=2e-16)
+        xp_assert_equal(res2, ref2)
+        xp_assert_equal(res3, ref3)
 
 
-@pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no expand_dims")
 class TestAtLeastND:
     def test_0D(self, xp: ModuleType):
         x = xp.asarray(1.0)
@@ -70,7 +289,7 @@ class TestAtLeastND:
         xp_assert_equal(y, xp.asarray([[0, 1]]))
 
         y = atleast_nd(x, ndim=5)
-        xp_assert_equal(y, xp.reshape(xp.arange(2), (1, 1, 1, 1, 2)))
+        xp_assert_equal(y, xp.asarray([[[[[0, 1]]]]]))
 
     def test_2D(self, xp: ModuleType):
         x = xp.asarray([[3.0]])
@@ -219,9 +438,11 @@ class TestCov:
         )
 
 
+@pytest.mark.skip_xp_backend(
+    Backend.SPARSE, reason="read-only backend without .at support"
+)
 class TestCreateDiagonal:
-    @pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no device kwarg in zeros()")
-    def test_1d(self, xp: ModuleType):
+    def test_1d_from_numpy(self, xp: ModuleType):
         # from np.diag tests
         vals = 100 * xp.arange(5, dtype=xp.float64)
         b = xp.zeros((5, 5), dtype=xp.float64)
@@ -236,10 +457,9 @@ class TestCreateDiagonal:
         xp_assert_equal(create_diagonal(vals, offset=2), b)
         xp_assert_equal(create_diagonal(vals, offset=-2), c)
 
-    @pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no device kwarg in zeros()")
     @pytest.mark.parametrize("n", range(1, 10))
     @pytest.mark.parametrize("offset", range(1, 10))
-    def test_create_diagonal(self, xp: ModuleType, n: int, offset: int):
+    def test_1d_from_scipy(self, xp: ModuleType, n: int, offset: int):
         # from scipy._lib tests
         rng = np.random.default_rng(2347823)
         one = xp.asarray(1.0)
@@ -248,20 +468,39 @@ class TestCreateDiagonal:
         B = xp.asarray(np.diag(x, offset), dtype=one.dtype)
         xp_assert_equal(A, B)
 
-    def test_0d(self, xp: ModuleType):
+    def test_0d_raises(self, xp: ModuleType):
         with pytest.raises(ValueError, match="1-dimensional"):
-            create_diagonal(xp.asarray(1))
+            _ = create_diagonal(xp.asarray(1))
 
-    def test_2d(self, xp: ModuleType):
-        with pytest.raises(ValueError, match="1-dimensional"):
-            create_diagonal(xp.asarray([[1]]))
+    @pytest.mark.parametrize(
+        "shape",
+        [
+            (0,),
+            (10,),
+            (0, 1),
+            (1, 0),
+            (0, 0),
+            (4, 2, 1),
+            (1, 1, 7),
+            (0, 0, 1),
+            (3, 2, 4, 5),
+        ],
+    )
+    def test_nd(self, xp: ModuleType, shape: tuple[int, ...]):
+        rng = np.random.default_rng(2347823)
+        b = xp.asarray(
+            rng.integers((1 << 64) - 1, size=shape, dtype=np.uint64), dtype=xp.uint64
+        )
+        c = create_diagonal(b)
+        zero = xp.zeros((), dtype=xp.uint64)
+        assert c.shape == (*b.shape, b.shape[-1])
+        for i in ndindex(*eager_shape(c)):
+            xp_assert_equal(c[i], b[i[:-1]] if i[-2] == i[-1] else zero)
 
-    @pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no device kwarg in zeros()")
     def test_device(self, xp: ModuleType, device: Device):
         x = xp.asarray([1, 2, 3], device=device)
         assert get_device(create_diagonal(x)) == device
 
-    @pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no device kwarg in zeros()")
     def test_xp(self, xp: ModuleType):
         x = xp.asarray([1, 2])
         y = create_diagonal(x, xp=xp)
@@ -269,25 +508,13 @@ class TestCreateDiagonal:
 
 
 class TestExpandDims:
-    @pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no expand_dims")
-    @pytest.mark.xfail_xp_backend(Backend.DASK, reason="tuple index out of range")
-    @pytest.mark.xfail_xp_backend(Backend.TORCH, reason="tuple index out of range")
-    def test_functionality(self, xp: ModuleType):
-        def _squeeze_all(b: Array) -> Array:
-            """Mimics `np.squeeze(b)`. `xpx.squeeze`?"""
-            for axis in range(b.ndim):
-                with contextlib.suppress(ValueError):
-                    b = xp.squeeze(b, axis=axis)
-            return b
-
-        s = (2, 3, 4, 5)
-        a = xp.empty(s)
+    def test_single_axis(self, xp: ModuleType):
+        """Trivial case where xpx.expand_dims doesn't add anything to xp.expand_dims"""
+        a = xp.empty((2, 3, 4, 5))
         for axis in range(-5, 4):
             b = expand_dims(a, axis=axis)
-            assert b.shape[axis] == 1
-            assert _squeeze_all(b).shape == s
+            xp_assert_equal(b, xp.expand_dims(a, axis=axis))
 
-    @pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no expand_dims")
     def test_axis_tuple(self, xp: ModuleType):
         a = xp.empty((3, 3, 3))
         assert expand_dims(a, axis=(0, 1, 2)).shape == (1, 1, 1, 3, 3, 3)
@@ -296,36 +523,33 @@ class TestExpandDims:
         assert expand_dims(a, axis=(0, -3, -5)).shape == (1, 1, 3, 1, 3, 3)
 
     def test_axis_out_of_range(self, xp: ModuleType):
-        s = (2, 3, 4, 5)
-        a = xp.empty(s)
+        a = xp.empty((2, 3, 4, 5))
         with pytest.raises(IndexError, match="out of bounds"):
-            expand_dims(a, axis=-6)
+            _ = expand_dims(a, axis=-6)
         with pytest.raises(IndexError, match="out of bounds"):
-            expand_dims(a, axis=5)
+            _ = expand_dims(a, axis=5)
 
         a = xp.empty((3, 3, 3))
         with pytest.raises(IndexError, match="out of bounds"):
-            expand_dims(a, axis=(0, -6))
+            _ = expand_dims(a, axis=(0, -6))
         with pytest.raises(IndexError, match="out of bounds"):
-            expand_dims(a, axis=(0, 5))
+            _ = expand_dims(a, axis=(0, 5))
 
     def test_repeated_axis(self, xp: ModuleType):
         a = xp.empty((3, 3, 3))
         with pytest.raises(ValueError, match="Duplicate dimensions"):
-            expand_dims(a, axis=(1, 1))
+            _ = expand_dims(a, axis=(1, 1))
 
     def test_positive_negative_repeated(self, xp: ModuleType):
         # https://github.com/data-apis/array-api/issues/760#issuecomment-1989449817
         a = xp.empty((2, 3, 4, 5))
         with pytest.raises(ValueError, match="Duplicate dimensions"):
-            expand_dims(a, axis=(3, -3))
+            _ = expand_dims(a, axis=(3, -3))
 
-    @pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no expand_dims")
     def test_device(self, xp: ModuleType, device: Device):
         x = xp.asarray([1, 2, 3], device=device)
         assert get_device(expand_dims(x, axis=0)) == device
 
-    @pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no expand_dims")
     def test_xp(self, xp: ModuleType):
         x = xp.asarray([1, 2, 3])
         y = expand_dims(x, axis=(0, 1, 2), xp=xp)
@@ -334,8 +558,6 @@ class TestExpandDims:
 
 @pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no isdtype")
 class TestIsClose:
-    # FIXME use lazywhere to avoid warnings on inf
-    @pytest.mark.filterwarnings("ignore:invalid value encountered")
     @pytest.mark.parametrize("swap", [False, True])
     @pytest.mark.parametrize(
         ("a", "b"),
@@ -394,8 +616,6 @@ class TestIsClose:
 
         xp_assert_equal(actual, expect)
 
-    # FIXME use lazywhere to avoid warnings on inf
-    @pytest.mark.filterwarnings("ignore:invalid value encountered")
     def test_some_inf(self, xp: ModuleType):
         a = xp.asarray([0.0, 1.0, xp.inf, xp.inf, xp.inf])
         b = xp.asarray([1e-9, 1.0, xp.inf, -xp.inf, 2.0])
@@ -484,7 +704,7 @@ class TestIsClose:
 
     def test_all_python_scalars(self):
         with pytest.raises(TypeError, match="Unrecognized"):
-            isclose(0, 0)
+            _ = isclose(0, 0)
 
     def test_xp(self, xp: ModuleType):
         a = xp.asarray([0.0, 0.0])
@@ -492,7 +712,6 @@ class TestIsClose:
         xp_assert_equal(isclose(a, b, xp=xp), xp.asarray([True, False]))
 
 
-@pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no expand_dims")
 class TestKron:
     def test_basic(self, xp: ModuleType):
         # Using 0-dimensional array
@@ -551,6 +770,7 @@ class TestKron:
         k = kron(a, b)
         assert k.shape == expected_shape
 
+    @pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no isdtype")
     def test_python_scalar(self, xp: ModuleType):
         a = 1
         # Test no dtype promotion to xp.asarray(a); use b.dtype
@@ -561,7 +781,7 @@ class TestKron:
 
     def test_all_python_scalars(self):
         with pytest.raises(TypeError, match="Unrecognized"):
-            kron(1, 1)
+            _ = kron(1, 1)
 
     def test_device(self, xp: ModuleType, device: Device):
         x1 = xp.asarray([1, 2, 3], device=device)
@@ -593,27 +813,29 @@ class TestNUnique:
         xp_assert_equal(nunique(a, xp=xp), xp.asarray(3))
 
 
-@pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no arange, no device")
 class TestPad:
     def test_simple(self, xp: ModuleType):
-        a = xp.arange(1, 4)
+        a = xp.asarray([1, 2, 3])
         padded = pad(a, 2)
         xp_assert_equal(padded, xp.asarray([0, 0, 1, 2, 3, 0, 0]))
 
+    @pytest.mark.xfail_xp_backend(
+        Backend.SPARSE, reason="constant_values can only be equal to fill value"
+    )
     def test_fill_value(self, xp: ModuleType):
-        a = xp.arange(1, 4)
+        a = xp.asarray([1, 2, 3])
         padded = pad(a, 2, constant_values=42)
         xp_assert_equal(padded, xp.asarray([42, 42, 1, 2, 3, 42, 42]))
 
     def test_ndim(self, xp: ModuleType):
-        a = xp.reshape(xp.arange(2 * 3 * 4), (2, 3, 4))
+        a = xp.asarray(np.reshape(np.arange(2 * 3 * 4), (2, 3, 4)))
         padded = pad(a, 2)
         assert padded.shape == (6, 7, 8)
 
     def test_mode_not_implemented(self, xp: ModuleType):
-        a = xp.arange(3)
+        a = xp.asarray([1, 2, 3])
         with pytest.raises(NotImplementedError, match="Only `'constant'`"):
-            pad(a, 2, mode="edge")  # type: ignore[arg-type]  # pyright: ignore[reportArgumentType]
+            _ = pad(a, 2, mode="edge")  # type: ignore[arg-type]  # pyright: ignore[reportArgumentType]
 
     def test_device(self, xp: ModuleType, device: Device):
         a = xp.asarray(0.0, device=device)
@@ -624,7 +846,7 @@ class TestPad:
         xp_assert_equal(padded, xp.asarray(0))
 
     def test_tuple_width(self, xp: ModuleType):
-        a = xp.reshape(xp.arange(12), (3, 4))
+        a = xp.asarray(np.reshape(np.arange(12), (3, 4)))
         padded = pad(a, (1, 0))
         assert padded.shape == (4, 5)
 
@@ -632,10 +854,10 @@ class TestPad:
         assert padded.shape == (6, 7)
 
         with pytest.raises((ValueError, RuntimeError)):
-            pad(a, [(1, 2, 3)])  # type: ignore[list-item]  # pyright: ignore[reportArgumentType]
+            _ = pad(a, [(1, 2, 3)])  # type: ignore[list-item]  # pyright: ignore[reportArgumentType]
 
     def test_sequence_of_tuples_width(self, xp: ModuleType):
-        a = xp.reshape(xp.arange(12), (3, 4))
+        a = xp.asarray(np.reshape(np.arange(12), (3, 4)))
 
         padded = pad(a, ((1, 0), (0, 2)))
         assert padded.shape == (4, 6)
@@ -657,7 +879,7 @@ assume_unique = pytest.mark.parametrize(
 )
 
 
-@pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no device kwarg in asarray()")
+@pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no argsort")
 class TestSetDiff1D:
     @pytest.mark.xfail_xp_backend(Backend.DASK, reason="NaN-shaped arrays")
     @pytest.mark.xfail_xp_backend(
@@ -724,7 +946,7 @@ class TestSetDiff1D:
     @pytest.mark.parametrize("assume_unique", [True, False])
     def test_all_python_scalars(self, assume_unique: bool):
         with pytest.raises(TypeError, match="Unrecognized"):
-            setdiff1d(0, 0, assume_unique=assume_unique)
+            _ = setdiff1d(0, 0, assume_unique=assume_unique)
 
     @assume_unique
     def test_device(self, xp: ModuleType, device: Device, assume_unique: bool):
@@ -752,7 +974,7 @@ class TestSinc:
     @pytest.mark.parametrize("x", [0, 1 + 3j])
     def test_dtype(self, xp: ModuleType, x: int | complex):
         with pytest.raises(ValueError, match="real floating data type"):
-            sinc(xp.asarray(x))
+            _ = sinc(xp.asarray(x))
 
     def test_3d(self, xp: ModuleType):
         x = xp.reshape(xp.arange(18, dtype=xp.float64), (3, 3, 2))
