@@ -1,5 +1,6 @@
 from collections.abc import Callable
 from types import ModuleType
+from typing import cast
 
 import numpy as np
 import pytest
@@ -14,7 +15,7 @@ from array_api_extra._lib._utils._compat import (
 from array_api_extra._lib._utils._typing import Array
 from array_api_extra.testing import lazy_xp_function
 
-# mypy: disable-error-code=no-any-decorated
+# mypy: disable-error-code=decorated-any
 # pyright: reportUnknownParameterType=false,reportMissingParameterType=false
 
 param_assert_equal_close = pytest.mark.parametrize(
@@ -30,7 +31,7 @@ param_assert_equal_close = pytest.mark.parametrize(
 
 
 @param_assert_equal_close
-def test_assert_close_equal_basic(xp: ModuleType, func: Callable[..., None]):  # type: ignore[no-any-explicit]
+def test_assert_close_equal_basic(xp: ModuleType, func: Callable[..., None]):  # type: ignore[explicit-any]
     func(xp.asarray(0), xp.asarray(0))
     func(xp.asarray([1, 2]), xp.asarray([1, 2]))
 
@@ -50,7 +51,7 @@ def test_assert_close_equal_basic(xp: ModuleType, func: Callable[..., None]):  #
 @pytest.mark.skip_xp_backend(Backend.NUMPY, reason="test other ns vs. numpy")
 @pytest.mark.skip_xp_backend(Backend.NUMPY_READONLY, reason="test other ns vs. numpy")
 @pytest.mark.parametrize("func", [xp_assert_equal, xp_assert_close])
-def test_assert_close_equal_namespace(xp: ModuleType, func: Callable[..., None]):  # type: ignore[no-any-explicit]
+def test_assert_close_equal_namespace(xp: ModuleType, func: Callable[..., None]):  # type: ignore[explicit-any]
     with pytest.raises(AssertionError, match="namespaces do not match"):
         func(xp.asarray(0), np.asarray(0))
     with pytest.raises(TypeError, match="Unrecognized array input"):
@@ -71,8 +72,8 @@ def test_assert_close_tolerance(xp: ModuleType):
 
 
 @param_assert_equal_close
-@pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="no bool indexing")
-def test_assert_close_equal_none_shape(xp: ModuleType, func: Callable[..., None]):  # type: ignore[no-any-explicit]
+@pytest.mark.xfail_xp_backend(Backend.SPARSE, reason="index by sparse array")
+def test_assert_close_equal_none_shape(xp: ModuleType, func: Callable[..., None]):  # type: ignore[explicit-any]
     """On dask and other lazy backends, test that a shape with NaN's or None's
     can be compared to a real shape.
     """
@@ -108,6 +109,7 @@ def non_materializable(x: Array) -> Array:
     and it will trigger an expensive computation in dask.
     """
     xp = array_namespace(x)
+    # Crashes inside jax.jit
     # On dask, this triggers two computations of the whole graph
     if xp.any(x < 0.0) or xp.any(x > 10.0):
         msg = "Values must be in the [0, 10] range"
@@ -150,19 +152,19 @@ def test_lazy_xp_function(xp: ModuleType):
         with pytest.raises(
             TypeError, match="Attempted boolean conversion of traced array"
         ):
-            non_materializable4(x)  # Wrapped
+            _ = non_materializable4(x)  # Wrapped
 
     elif is_dask_namespace(xp):
         with pytest.raises(
             AssertionError,
             match=r"dask\.compute.* 2 times, but only up to 1 calls are allowed",
         ):
-            non_materializable3(x)
+            _ = non_materializable3(x)
         with pytest.raises(
             AssertionError,
             match=r"dask\.compute.* 1 times, but no calls are allowed",
         ):
-            non_materializable4(x)
+            _ = non_materializable4(x)
 
     else:
         xp_assert_equal(non_materializable3(x), xp.asarray([1.0, 2.0]))
@@ -194,7 +196,7 @@ lazy_xp_function(static_params3, static_argnums=1, static_argnames="flag")
 
 
 @pytest.mark.parametrize("func", [static_params1, static_params2, static_params3])
-def test_lazy_xp_function_static_params(xp: ModuleType, func: Callable[..., Array]):  # type: ignore[no-any-explicit]
+def test_lazy_xp_function_static_params(xp: ModuleType, func: Callable[..., Array]):  # type: ignore[explicit-any]
     x = xp.asarray([1.0, 2.0])
     xp_assert_equal(func(x, 1), xp.asarray([3.0, 6.0]))
     xp_assert_equal(func(x, 1, True), xp.asarray([2.0, 4.0]))
@@ -202,3 +204,102 @@ def test_lazy_xp_function_static_params(xp: ModuleType, func: Callable[..., Arra
     xp_assert_equal(func(x, 0, False), xp.asarray([3.0, 6.0]))
     xp_assert_equal(func(x, 1, flag=True), xp.asarray([2.0, 4.0]))
     xp_assert_equal(func(x, n=1, flag=True), xp.asarray([2.0, 4.0]))
+
+
+try:
+    # Test an arbitrary Cython ufunc (@cython.vectorize).
+    # When SCIPY_ARRAY_API is not set, this is the same as
+    # scipy.special.erf.
+    from scipy.special._ufuncs import erf  # type: ignore[import-not-found]
+
+    lazy_xp_function(erf)  # pyright: ignore[reportUnknownArgumentType]
+except ImportError:
+    erf = None
+
+
+@pytest.mark.filterwarnings("ignore:__array_wrap__:DeprecationWarning")  # torch
+def test_lazy_xp_function_cython_ufuncs(xp: ModuleType, library: Backend):
+    pytest.importorskip("scipy")
+    assert erf is not None
+    x = xp.asarray([6.0, 7.0])
+    if library in (Backend.ARRAY_API_STRICT, Backend.JAX):
+        # array-api-strict arrays are auto-converted to numpy
+        # which results in an assertion error for mismatched namespaces
+        # eager jax arrays are auto-converted to numpy in eager jax
+        # and fail in jax.jit (which lazy_xp_function tests here)
+        with pytest.raises((TypeError, AssertionError)):
+            xp_assert_equal(cast(Array, erf(x)), xp.asarray([1.0, 1.0]))
+    else:
+        # cupy, dask and sparse define __array_ufunc__ and dispatch accordingly
+        # note that when sparse reduces to scalar it returns a np.generic, which
+        # would make xp_assert_equal fail.
+        xp_assert_equal(cast(Array, erf(x)), xp.asarray([1.0, 1.0]))
+
+
+def dask_raises(x: Array) -> Array:
+    def _raises(x: Array) -> Array:
+        # Test that map_blocks doesn't eagerly call the function;
+        # dtype and meta should be sufficient to skip the trial run.
+        assert x.shape == (3,)
+        msg = "Hello world"
+        raise ValueError(msg)
+
+    return x.map_blocks(_raises, dtype=x.dtype, meta=x._meta)  # type: ignore[attr-defined]  # pyright: ignore[reportAttributeAccessIssue]
+
+
+lazy_xp_function(dask_raises)
+
+
+def test_lazy_xp_function_eagerly_raises(da: ModuleType):
+    """Test that the pattern::
+
+        with pytest.raises(Exception):
+            func(x)
+
+    works with Dask, even though it normally wouldn't as we're disregarding the func
+    output so the graph would not be ordinarily materialized.
+    lazy_xp_function contains ad-hoc code to materialize and reraise exceptions.
+    """
+    x = da.arange(3)
+    with pytest.raises(ValueError, match="Hello world"):
+        _ = dask_raises(x)
+
+
+wrapped = ModuleType("wrapped")
+naked = ModuleType("naked")
+
+
+def f(x: Array) -> Array:
+    xp = array_namespace(x)
+    # Crash in jax.jit and trigger compute() on dask
+    if not xp.all(x):
+        msg = "Values must be non-zero"
+        raise ValueError(msg)
+    return x
+
+
+wrapped.f = f  # type: ignore[attr-defined]  # pyright: ignore[reportAttributeAccessIssue]
+naked.f = f  # type: ignore[attr-defined]  # pyright: ignore[reportAttributeAccessIssue]
+del f
+
+
+lazy_xp_function(wrapped.f)
+lazy_xp_modules = [wrapped]
+
+
+def test_lazy_xp_modules(xp: ModuleType, library: Backend):
+    x = xp.asarray([1.0, 2.0])
+    y = naked.f(x)
+    xp_assert_equal(y, x)
+
+    if library is Backend.JAX:
+        with pytest.raises(
+            TypeError, match="Attempted boolean conversion of traced array"
+        ):
+            wrapped.f(x)
+    elif library is Backend.DASK:
+        with pytest.raises(AssertionError, match=r"dask\.compute"):
+            wrapped.f(x)
+    else:
+        y = wrapped.f(x)
+        xp_assert_equal(y, x)
